# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later


"""Data loading code."""

import pathlib

import numpy
import PIL.Image


class SingleAutoLevel16to8:
    """Converts a 16-bit image to 8-bit representation using "auto-level".

    This transform assumes that the input image is gray-scaled.

    To auto-level, we calculate the maximum and the minimum of the image, and
    consider such a range should be mapped to the [0,255] range of the
    destination image.
    """

    def __call__(self, img):
        imin, imax = img.getextrema()
        irange = imax - imin
        return PIL.Image.fromarray(
            numpy.round(
                255.0 * (numpy.array(img).astype(float) - imin) / irange
            ).astype("uint8"),
        ).convert("L")


class RemoveBlackBorders:
    """Remove black borders of CXR."""

    def __init__(self, threshold=0):
        self.threshold = threshold

    def __call__(self, img):
        img = numpy.asarray(img)
        mask = numpy.asarray(img) > self.threshold
        return PIL.Image.fromarray(img[numpy.ix_(mask.any(1), mask.any(0))])


def load_pil(path: str | pathlib.Path) -> PIL.Image.Image:
    """Loads a sample data.

    Parameters
    ----------

    path
        The full path leading to the image to be loaded


    Returns
    -------

    image
        A PIL image
    """
    return PIL.Image.open(path)


def load_pil_baw(path: str | pathlib.Path) -> PIL.Image.Image:
    """Loads a sample data.

    Parameters
    ----------

    path
        The full path leading to the image to be loaded


    Returns
    -------

    image
        A PIL image in grayscale mode
    """
    return load_pil(path).convert("L")


def load_pil_rgb(path: str | pathlib.Path) -> PIL.Image.Image:
    """Loads a sample data.

    Parameters
    ----------

    path
        The full path leading to the image to be loaded


    Returns
    -------

    image
        A PIL image in RGB mode
    """
    return load_pil(path).convert("RGB")
