# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Shenzhen dataset for computer-aided diagnosis.

The standard digital image database for Tuberculosis is created by the National
Library of Medicine, Maryland, USA in collaboration with Shenzhen No.3 People’s
Hospital, Guangdong Medical College, Shenzhen, China. The Chest X-rays are from
out-patient clinics, and were captured as part of the daily routine using
Philips DR Digital Diagnose systems.

* Reference: [MONTGOMERY-SHENZHEN-2014]_
* Original resolution (height x width or width x height): 3000 x 3000 or less
* Split reference: none
* Protocol ``default``:

  * Training samples: 64% of TB and healthy CXR (including labels)
  * Validation samples: 16% of TB and healthy CXR (including labels)
  * Test samples: 20% of TB and healthy CXR (including labels)
"""

import os

import torchvision.transforms

from ...utils.rc import load_rc
from ..image_utils import RemoveBlackBorders, load_pil_baw
from ..typing import RawDataLoader as _BaseRawDataLoader
from ..typing import Sample


class RawDataLoader(_BaseRawDataLoader):
    """A specialized raw-data-loader for the Shenzen dataset.

    Attributes
    ----------

    datadir
        This variable contains the base directory where the database raw data
        is stored.

    transform
        Transforms that are always applied to the loaded raw images.
    """

    datadir: str
    transform: torchvision.transforms.Compose

    def __init__(self):
        self.datadir = load_rc().get(
            "datadir.shenzhen", os.path.realpath(os.curdir)
        )

        self.transform = torchvision.transforms.Compose(
            [
                RemoveBlackBorders(),
                torchvision.transforms.Resize(512),
                torchvision.transforms.CenterCrop(512),
                torchvision.transforms.ToTensor(),
            ]
        )

    def sample(self, sample: tuple[str, int]) -> Sample:
        """Loads a single image sample from the disk.

        Parameters
        ----------

        sample:
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing the
            sample label.


        Returns
        -------

        sample
            The sample representation
        """
        tensor = self.transform(
            load_pil_baw(os.path.join(self.datadir, sample[0]))
        )
        return tensor, dict(label=sample[1])  # type: ignore[arg-type]

    def label(self, sample: tuple[str, int]) -> int:
        """Loads a single image sample label from the disk.

        Parameters
        ----------

        sample:
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing the
            sample label.


        Returns
        -------

        label
            The integer label associated with the sample
        """
        return sample[1]
