# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import csv
import importlib.abc
import json
import logging
import pathlib
import typing

import torch

from .typing import DatabaseSplit, RawDataLoader

logger = logging.getLogger(__name__)


class JSONDatabaseSplit(DatabaseSplit):
    """Defines a loader that understands a database split (train, test, etc) in
    JSON format.

    To create a new database split, you need to provide a JSON formatted
    dictionary in a file, with contents similar to the following:

    .. code-block:: json

       {
           "subset1": [
               [
                   "sample1-data1",
                   "sample1-data2",
                   "sample1-data3",
               ],
               [
                   "sample2-data1",
                   "sample2-data2",
                   "sample2-data3",
               ]
           ],
           "subset2": [
               [
                   "sample42-data1",
                   "sample42-data2",
                   "sample42-data3",
               ],
           ]
       }

    Your database split many contain any number of subsets (dictionary keys).
    For simplicity, we recommend all sample entries are formatted similarly so
    that raw-data-loading is simplified.  Use the function
    :py:func:`check_database_split_loading` to test raw data loading and fine
    tune the dataset split, or its loading.

    Objects of this class behave like a dictionary in which keys are subset
    names in the split, and values represent samples data and meta-data.


    Parameters
    ----------

    path
        Absolute path to a JSON formatted file containing the database split to be
        recognized by this object.
    """

    def __init__(self, path: pathlib.Path | str | importlib.abc.Traversable):
        if isinstance(path, str):
            path = pathlib.Path(path)
        self.path = path
        self.subsets = self._load_split_from_disk()

    def _load_split_from_disk(self) -> DatabaseSplit:
        """Loads all subsets in a split from its file system representation.

        This method will load JSON information for the current split and return
        all subsets of the given split after converting each entry through the
        loader function.


        Returns
        -------

        subsets : dict
            A dictionary mapping subset names to lists of JSON objects
        """

        if str(self.path).endswith(".bz2"):
            logger.debug(f"Loading database split from {str(self.path)}...")
            with __import__("bz2").open(self.path) as f:
                return json.load(f)
        else:
            with self.path.open() as f:
                return json.load(f)

    def __getitem__(self, key: str) -> typing.Sequence[typing.Any]:
        """Accesses subset ``key`` from this split."""
        return self.subsets[key]

    def __iter__(self):
        """Iterates over the subsets."""
        return iter(self.subsets)

    def __len__(self) -> int:
        """How many subsets we currently have."""
        return len(self.subsets)


class CSVDatabaseSplit(DatabaseSplit):
    """Defines a loader that understands a database split (train, test, etc) in
    CSV format.

    To create a new database split, you need to provide one or more CSV
    formatted files, each representing a subset of this split, containing the
    sample data (one per row).  Example:

    Inside the directory ``my-split/``, one can file files ``train.csv``,
    ``validation.csv``, and ``test.csv``.  Each file has a structure similar to
    the following:

    .. code-block:: text

       sample1-value1,sample1-value2,sample1-value3
       sample2-value1,sample2-value2,sample2-value3
       ...

    Each file in the provided directory defines the subset name on the split.
    So, the file ``train.csv`` will contain the data from the ``train`` subset,
    and so on.

    Objects of this class behave like a dictionary in which keys are subset
    names in the split, and values represent samples data and meta-data.


    Parameters
    ----------

    directory
        Absolute path to a directory containing the database split layed down
        as a set of CSV files, one per subset.
    """

    def __init__(
        self, directory: pathlib.Path | str | importlib.abc.Traversable
    ):
        if isinstance(directory, str):
            directory = pathlib.Path(directory)
        assert (
            directory.is_dir()
        ), f"`{str(directory)}` is not a valid directory"
        self.directory = directory
        self.subsets = self._load_split_from_disk()

    def _load_split_from_disk(self) -> DatabaseSplit:
        """Loads all subsets in a split from its file system representation.

        This method will load CSV information for the current split and return all
        subsets of the given split after converting each entry through the
        loader function.


        Returns
        -------

        subsets : dict
            A dictionary mapping subset names to lists of JSON objects
        """

        retval: DatabaseSplit = {}
        for subset in self.directory.iterdir():
            if str(subset).endswith(".csv.bz2"):
                logger.debug(f"Loading database split from {subset}...")
                with __import__("bz2").open(subset) as f:
                    reader = csv.reader(f)
                    retval[subset.name[: -len(".csv.bz2")]] = [
                        k for k in reader
                    ]
            elif str(subset).endswith(".csv"):
                with subset.open() as f:
                    reader = csv.reader(f)
                    retval[subset.name[: -len(".csv")]] = [k for k in reader]
            else:
                logger.debug(
                    f"Ignoring file {subset} in CSVDatabaseSplit readout"
                )
        return retval

    def __getitem__(self, key: str) -> typing.Sequence[typing.Any]:
        """Accesses subset ``key`` from this split."""
        return self.subsets[key]

    def __iter__(self):
        """Iterates over the subsets."""
        return iter(self.subsets)

    def __len__(self) -> int:
        """How many subsets we currently have."""
        return len(self.subsets)


def check_database_split_loading(
    database_split: DatabaseSplit,
    loader: RawDataLoader,
    limit: int = 0,
) -> int:
    """For each subset in the split, check if all data can be correctly loaded
    using the provided loader function.

    This function will return the number of errors loading samples, and will
    log more detailed information to the logging stream.


    Parameters
    ----------

    database_split
        A mapping that, contains the database split.  Each key represents the
        name of a subset in the split.  Each value is a (potentially complex)
        object that represents a single sample.

    loader
        A loader object that knows how to handle full-samples or just labels.

    limit
        Maximum number of samples to check (in each split/subset
        combination) in this dataset.  If set to zero, then check
        everything.


    Returns
    -------

    errors
        Number of errors found
    """
    logger.info(
        "Checking if can load all samples in all subsets of this split..."
    )
    errors = 0
    for subset in database_split.keys():
        samples = subset if not limit else subset[:limit]
        for pos, sample in enumerate(samples):
            try:
                data, _ = loader.sample(sample)
                assert isinstance(data, torch.Tensor)
            except Exception as e:
                logger.info(
                    f"Found error loading entry {pos} in subset `{subset}`: {e}"
                )
                errors += 1
    return errors
