# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Image transformations for our pipelines.

Differences between methods here and those from
:py:mod:`torchvision.transforms` is that these support multiple
simultaneous image inputs, which are required to feed segmentation
networks (e.g. image and labels or masks).  We also take care of data
augmentations, in which random flipping and rotation needs to be applied
across all input images, but color jittering, for example, only on the
input image.
"""

import random

import numpy
import PIL.Image

from scipy.ndimage import gaussian_filter, map_coordinates
from torchvision import transforms


class ElasticDeformation:
    """Elastic deformation of 2D image slightly adapted from [SIMARD-2003]_.

    Source: https://gist.github.com/oeway/2e3b989e0343f0884388ed7ed82eb3b0
    """

    def __init__(
        self,
        alpha=1000,
        sigma=30,
        spline_order=1,
        mode="nearest",
        random_state=numpy.random,
        p=1.0,
    ):
        self.alpha = alpha
        self.sigma = sigma
        self.spline_order = spline_order
        self.mode = mode
        self.random_state = random_state
        self.p = p

    def __call__(self, img):
        if random.random() < self.p:
            assert img.ndim == 3

            # Input tensor is of shape C x H x W
            # If the tensor only contains one channel, this conversion results in H x W.
            # With 3 channels, we get H x W x C
            img = transforms.ToPILImage()(img)
            img = numpy.asarray(img)

            shape = img.shape[:2]

            dx = (
                gaussian_filter(
                    (self.random_state.rand(*shape) * 2 - 1),
                    self.sigma,
                    mode="constant",
                    cval=0,
                )
                * self.alpha
            )
            dy = (
                gaussian_filter(
                    (self.random_state.rand(*shape) * 2 - 1),
                    self.sigma,
                    mode="constant",
                    cval=0,
                )
                * self.alpha
            )

            x, y = numpy.meshgrid(
                numpy.arange(shape[0]), numpy.arange(shape[1]), indexing="ij"
            )
            indices = [
                numpy.reshape(x + dx, (-1, 1)),
                numpy.reshape(y + dy, (-1, 1)),
            ]
            result = numpy.empty_like(img)

            if img.ndim == 2:
                result[:, :] = map_coordinates(
                    img[:, :], indices, order=self.spline_order, mode=self.mode
                ).reshape(shape)

            else:
                for i in range(img.shape[2]):
                    result[:, :, i] = map_coordinates(
                        img[:, :, i],
                        indices,
                        order=self.spline_order,
                        mode=self.mode,
                    ).reshape(shape)

            return transforms.ToTensor()(PIL.Image.fromarray(result))

        else:
            return img
