# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import csv
import logging
import os
import shutil

import lightning.pytorch
import lightning.pytorch.callbacks
import lightning.pytorch.loggers
import torch.nn

from ..utils.accelerator import AcceleratorProcessor
from ..utils.resources import ResourceMonitor, cpu_constants, gpu_constants
from .callbacks import LoggingCallback

logger = logging.getLogger(__name__)


def save_model_summary(
    output_folder: str, model: torch.nn.Module
) -> tuple[lightning.pytorch.callbacks.ModelSummary, int]:
    """Save a little summary of the model in a txt file.

    Parameters
    ----------

    output_folder
        output path

    model
        Network (e.g. driu, hed, unet)

    Returns
    -------
    summary:
        The model summary in a text format.

    total_parameters:
        The number of parameters of the model.
    """
    summary_path = os.path.join(output_folder, "model_summary.txt")
    logger.info(f"Saving model summary at {summary_path}...")
    with open(summary_path, "w") as f:
        summary = lightning.pytorch.utilities.model_summary.ModelSummary(
            model, max_depth=-1
        )
        f.write(str(summary))
    return (
        summary,
        lightning.pytorch.utilities.model_summary.ModelSummary(
            model
        ).total_parameters,
    )


def static_information_to_csv(static_logfile_name, device, n):
    """Save the static information in a csv file.

    Parameters
    ----------

    static_logfile_name : str
        The static file name which is a join between the output folder and
        "constant.csv"
    """
    if os.path.exists(static_logfile_name):
        backup = static_logfile_name + "~"
        if os.path.exists(backup):
            os.unlink(backup)
        shutil.move(static_logfile_name, backup)
    with open(static_logfile_name, "w", newline="") as f:
        logdata: dict[str, int | float | str] = {}
        logdata.update(cpu_constants())
        if device == "cuda":
            results = gpu_constants()
            if results is not None:
                logdata.update(results)
        logdata["model_size"] = n
        logwriter = csv.DictWriter(f, fieldnames=logdata.keys())
        logwriter.writeheader()
        logwriter.writerow(logdata)


def run(
    model,
    datamodule,
    checkpoint_period,
    accelerator,
    arguments,
    output_folder,
    monitoring_interval,
    batch_chunk_count,
    checkpoint,
):
    """Fits a CNN model using supervised learning and save it to disk.

    This method supports periodic checkpointing and the output of a
    CSV-formatted log with the evolution of some figures during training.


    Parameters
    ----------

    model : :py:class:`torch.nn.Module`
        Neural network model (e.g. pasa).

    data_loader : :py:class:`torch.utils.data.DataLoader`
        The pytorch Dataloader used to iterate over batches.

    valid_loaders : :py:class:`list` of :py:class:`torch.utils.data.DataLoader`
        To be used to validate the model and enable automatic checkpointing.
        If ``None``, then do not validate it.

    extra_valid_loaders : :py:class:`list` of :py:class:`torch.utils.data.DataLoader`
        To be used to validate the model, however **does not affect** automatic
        checkpointing. If empty, then does not log anything else.  Otherwise,
        an extra column with the loss of every dataset in this list is kept on
        the final training log.

    checkpoint_period : int
        Save a checkpoint every ``n`` epochs.  If set to ``0`` (zero), then do
        not save intermediary checkpoints.

    accelerator : str
        A string indicating the accelerator to use (e.g. "cpu" or "gpu"). The
        device can also be specified (gpu:0).

    arguments : dict
        Start and end epochs:

    output_folder : str
        Directory in which the results will be saved.

    monitoring_interval : int, float
        Interval, in seconds (or fractions), through which we should monitor
        resources during training.

    batch_chunk_count: int
        If this number is different than 1, then each batch will be divided in
        this number of chunks.  Gradients will be accumulated to perform each
        mini-batch.   This is particularly interesting when one has limited RAM
        on the GPU, but would like to keep training with larger batches.  One
        exchanges for longer processing times in this case.
    """

    max_epoch = arguments["max_epoch"]

    accelerator_processor = AcceleratorProcessor(accelerator)

    os.makedirs(output_folder, exist_ok=True)

    # Save model summary
    _, no_of_parameters = save_model_summary(output_folder, model)

    csv_logger = lightning.pytorch.loggers.CSVLogger(output_folder, "logs_csv")
    tensorboard_logger = lightning.pytorch.loggers.TensorBoardLogger(
        output_folder, "logs_tensorboard"
    )

    resource_monitor = ResourceMonitor(
        interval=monitoring_interval,
        has_gpu=(accelerator_processor.accelerator == "gpu"),
        main_pid=os.getpid(),
        logging_level=logging.ERROR,
    )

    checkpoint_callback = lightning.pytorch.callbacks.ModelCheckpoint(
        output_folder,
        "model_lowest_valid_loss",
        save_last=True,
        monitor="validation_loss",
        mode="min",
        save_on_train_epoch_end=True,
        every_n_epochs=checkpoint_period,
    )

    checkpoint_callback.CHECKPOINT_NAME_LAST = "model_final_epoch"

    # write static information to a CSV file
    static_logfile_name = os.path.join(output_folder, "constants.csv")
    static_information_to_csv(
        static_logfile_name,
        accelerator_processor.to_torch(),
        no_of_parameters,
    )

    with resource_monitor:
        trainer = lightning.pytorch.Trainer(
            accelerator=accelerator_processor.accelerator,
            devices=(accelerator_processor.device or "auto"),
            max_epochs=max_epoch,
            accumulate_grad_batches=batch_chunk_count,
            logger=[csv_logger, tensorboard_logger],
            check_val_every_n_epoch=1,
            callbacks=[LoggingCallback(resource_monitor), checkpoint_callback],
        )

        _ = trainer.fit(model, datamodule, ckpt_path=checkpoint)
