# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging

import lightning.pytorch as pl
import torch
import torch.nn as nn
import torchvision.models as models
import torchvision.transforms

logger = logging.getLogger(__name__)


class Densenet(pl.LightningModule):
    """Densenet module.

    Note: only usable with a normalized dataset
    """

    def __init__(
        self,
        criterion=None,
        criterion_valid=None,
        optimizer=None,
        optimizer_configs=None,
        pretrained=False,
        augmentation_transforms=[],
    ):
        super().__init__()

        self.name = "Densenet"

        self.augmentation_transforms = torchvision.transforms.Compose(
            augmentation_transforms
        )

        self.criterion = criterion
        self.criterion_valid = criterion_valid

        self.optimizer = optimizer
        self.optimizer_configs = optimizer_configs

        self.normalizer = None
        self.pretrained = pretrained

        # Load pretrained model
        if not pretrained:
            weights = None
        else:
            logger.info("Loading pretrained model weights")
            weights = models.DenseNet121_Weights.DEFAULT

        self.model_ft = models.densenet121(weights=weights)

        # Adapt output features
        self.model_ft.classifier = nn.Sequential(
            nn.Linear(1024, 256), nn.Linear(256, 1)
        )

    def forward(self, x):
        x = self.normalizer(x)
        x = self.model_ft(x)

        return x

    def set_normalizer(self, dataloader: torch.utils.data.DataLoader) -> None:
        """Initializes the normalizer for the current model.

        This function is NOOP if ``pretrained = True`` (normalizer set to
        imagenet weights, during contruction).

        Parameters
        ----------

        dataloader: :py:class:`torch.utils.data.DataLoader`
            A torch Dataloader from which to compute the mean and std.
            Will not be used if the model is pretrained.
        """
        if self.pretrained:
            from .normalizer import make_imagenet_normalizer

            logger.warning(
                "ImageNet pre-trained densenet model - NOT "
                "computing z-norm factors from training data. "
                "Using preset factors from torchvision."
            )
            self.normalizer = make_imagenet_normalizer()
        else:
            from .normalizer import make_z_normalizer

            logger.info(
                "Uninitialised densenet model - "
                "computing z-norm factors from training data."
            )
            self.normalizer = make_z_normalizer(dataloader)

    def set_bce_loss_weights(self, datamodule):
        """Reweights loss weights if BCEWithLogitsLoss is used.

        Parameters
        ----------

        datamodule:
            A datamodule implementing train_dataloader() and val_dataloader()
        """
        from ..data.dataset import _get_positive_weights

        if isinstance(self.criterion, torch.nn.BCEWithLogitsLoss):
            logger.info("Reweighting BCEWithLogitsLoss training criterion.")
            train_positive_weights = _get_positive_weights(
                datamodule.train_dataloader()
            )
            self.criterion = torch.nn.BCEWithLogitsLoss(
                pos_weight=train_positive_weights
            )

        if isinstance(self.criterion_valid, torch.nn.BCEWithLogitsLoss):
            logger.info("Reweighting BCEWithLogitsLoss validation criterion.")
            validation_positive_weights = _get_positive_weights(
                datamodule.val_dataloader()["validation"]
            )
            self.criterion_valid = torch.nn.BCEWithLogitsLoss(
                pos_weight=validation_positive_weights
            )

    def training_step(self, batch, batch_idx):
        images = batch[0]
        labels = batch[1]["label"]

        # Increase label dimension if too low
        # Allows single and multiclass usage
        if labels.ndim == 1:
            labels = torch.reshape(labels, (labels.shape[0], 1))

        # Forward pass on the network
        augmented_images = [
            self.augmentation_transforms(img).to(self.device) for img in images
        ]
        # Combine list of augmented images back into a tensor
        augmented_images = torch.cat(augmented_images, 0).view(images.shape)
        outputs = self(augmented_images)

        training_loss = self.criterion(outputs, labels.float())

        return {"loss": training_loss}

    def validation_step(self, batch, batch_idx, dataloader_idx=0):
        images = batch[0]
        labels = batch[1]["label"]

        # Increase label dimension if too low
        # Allows single and multiclass usage
        if labels.ndim == 1:
            labels = torch.reshape(labels, (labels.shape[0], 1))

        # data forwarding on the existing network
        outputs = self(images)

        validation_loss = self.criterion_valid(outputs, labels.float())

        if dataloader_idx == 0:
            return {"validation_loss": validation_loss}
        else:
            return {f"extra_validation_loss_{dataloader_idx}": validation_loss}

    def predict_step(self, batch, batch_idx, dataloader_idx=0, grad_cams=False):
        images = batch[0]
        labels = batch[1]["label"]
        names = batch[1]["name"]

        outputs = self(images)
        probabilities = torch.sigmoid(outputs)

        # necessary check for HED architecture that uses several outputs
        # for loss calculation instead of just the last concatfuse block
        if isinstance(outputs, list):
            outputs = outputs[-1]

        return names[0], torch.flatten(probabilities), torch.flatten(labels)

    def configure_optimizers(self):
        optimizer = self.optimizer(self.parameters(), **self.optimizer_configs)
        return optimizer
